/**
* Copyright Since 2005 Ortus Solutions, Corp
* www.ortussolutions.com
* ---
* Provides our magical PDF tools based on the Jars included in Lucee already
* Dependencies: com.lowagie.text.pdf.*
*/
component accessors="true"{

	// Location of the PDF lib path
	variables.LIB_PATH = "lib/pdfbox-app-2.0.0-RC3.jar";

	/******************************** PUBLIC PROPERTIES ********************************/

	// Tag Attributes
	this.metadata.attributes = {
		action 			= { required:false,	type:"string",	"default":"read" },
		result			= { required:false,	type:"string",	"default":"pdfdata" },
		source 			= { required:true, 	type:"string" },
		destination 	= { required:false,	type:"string",	"default":"" },
		flatten 		= { required:false, type:"boolean", "default": "true" },
		xmlData 		= { required:false, type:"string" },
		fdfData 		= { required:false, type:"string" },
		overwriteData 	= { required:false, type="boolean", "default": "false" },
		structData 		= { required:false, type:"struct" },
		jsonData 		= { required:false, type:"string" }
	};

	/******************************** PROPERTIES ********************************/

	property name="hasEndTag" 	type="boolean";
	property name="formParams"	type="struct";
	property name="caller"		type="struct";

	/******************************** PUBLIC METHODS ********************************/

	/**
	* Constructor
	* @hasEndTag Do we have the end tag
	* @caller The caller binding scope
	*/
	OrtusPDF function init( required boolean hasEndTag, struct caller ){
		variables.hasEndTag 	= arguments.hasEndTag;
		variables.formParams 	= {};
		variables.caller 		= arguments.caller;

		return this;
	}

	/**
	* Start tag execution
	* @attributes Tag attributes
	* @caller Caller
	*/
	boolean function onStartTag( struct attributes ){
		variables.formparams = {};
		if( NOT variables.hasEndTag ){
			onEndTag( arguments.attributes, "" );
		}
		return variables.hasEndTag;
	}

	/**
	* End tag execution
	* @attributes Tag attributes
	* @generatedContent The generated content
	*/
	boolean function onEndTag( struct attributes, string generatedContent ){
		// Call action method
		var results = callAction( attributes.action, attributes );

		return true;
	}

	/**
	* Set form params
	* @name The name of the key in the form params
	* @value The value
	*/
	OrtusPDF function setFormParam( required string name, required string value ){
		variables.formparams[ arguments.name ] = arguments.value;
		return this;
	}

	/**
	* Dynamic method calling
	*/
	any function onMissingMethod( required missingMethodName, required missingMethodArguments ){
		return callAction( missingMethodName, missingMethodArguments );
	}

	/**
	* Call Action methods, we prefix them with _ and are private
	* @methodName The method name to execute
	* @args The arguments for the method
	*/
	any function callAction( required string methodName, required struct args ){
		var methodPointer = this[ "_" & methodName ];
		return methodPointer( argumentCollection=args );
    }

	/******************************** PRIVATE METHODS ********************************/

	/**
	* Read action
	* @source The source PDF
	* @result The results variable name or file location
	* @xmlData If passed, this is the name of the variable to output the fields as XML data
	* @fdfData If passed, this is the name of the file location to export the fields in FDF data
	* @jsonData If passed, this is the name of the variable to output the fileds as JSON data
	*/
	private OrtusPDF function _read(
		required string source,
		string result,
		string xmlData,
		string fdfData,
		string jsonData
	){
		// FDF Data exporting
		if( structKeyExists( arguments, "fdfData" ) ){
			return generateFDFData( arguments.source, arguments.fdfData );
		}

		// Get form fields
		var returnFields = getPDFFormFields( arguments.source );

    	// Check if incoming result is a file location ending with XML
    	if( right( arguments.result, 4 ) == ".xml" ){
    		fileWrite( arguments.result, generateXMLData( returnFields ) );
    	} else {
    		// Setup result on caller according to results attribute
			variables.caller[ arguments.result ] = returnFields;
    	}

    	// XML Data Output?
    	if( !isNull( arguments.xmlData ) ){
    		variables.caller[ arguments.xmlData ] = generateXMLData( returnFields );
    	}

    	// JSON Data Output?
    	if( !isNull( arguments.jsonData ) ){
    		variables.caller[ arguments.jsonData ] = serializeJSON( returnFields );
    	}

		return this;
	}

	/**
	* Generate FDF data from an incoming source file
	* @source The source file
	* @destination The destination file
	*/
	private OrtusPDF function generateFDFData( required string source, required string destination ){
		// validate source
		if ( !fileExists( arguments.source ) ){
			throw( type="OrtusPDF.FileNotFound", message="The file '#arguments.source#' could not be found" );
		}

		// java helpers
		var fileInputStream 	= createObject( "java", "java.io.FileInputStream" ).init( arguments.source );
		var PDFReader 			= createObject( "java", "org.apache.pdfbox.pdmodel.PDDocument", variables.LIB_PATH ).load( fileInputStream );

		try{
			// Load source and document catalog
			var catalog 	= PDFReader.getDocumentCatalog();
			var acroForm 	= catalog.getAcroForm();
			// export to file
			acroForm.exportFDF().save( arguments.destination );
		}
	    catch( any e ){
	    	rethrow;
    	} finally{
    		// close it up
    		fileInputStream.close();
			PDFReader.close();
    	}

    	return this;
	}

	/**
	 * Get PDF form fields as a native struct
	 *
	 * @source The source file
	 */
	private struct function getPDFFormFields( required string source ){
		// validate source
		if ( !fileExists( arguments.source ) ){
			throw( type="OrtusPDF.FileNotFound", message="The file '#arguments.source#' could not be found" );
		}

		// java helpers
		var fileInputStream = createObject( "java", "java.io.FileInputStream" ).init( arguments.source );
		var PDFReader 		= createObject( "java", "org.apache.pdfbox.pdmodel.PDDocument", variables.LIB_PATH ).load( fileInputStream );

		try{
			// Load source and document catalog
			var catalog = PDFReader.getDocumentCatalog();

			// Recursive processing of fields
			var processFields = function( fields, accumulator ){
				for( var thisField in arguments.fields ){
					if( thisField.getClass().getName().findNoCase( "org.apache.pdfbox.pdmodel.interactive.form.PDNonTerminalField" ) ){
						processFields( thisField.getChildren(), arguments.accumulator );
					} else {
						arguments.accumulator[ thisField.getFullyQualifiedName() ] = thisField.getValueAsString();
					}
				}
				return arguments.accumulator;
			};

			// Get array of fields returns List<PDField>
			// https://pdfbox.apache.org/docs/2.0.0-SNAPSHOT/javadocs/org/apache/pdfbox/pdmodel/interactive/form/PDField.html
			var returnFields = processFields( catalog.getAcroForm().getFields(), {} );
		}
	    catch( any e ){
	    	rethrow;
    	} finally{
    		// close it up
    		fileInputStream.close();
			PDFReader.close();
    	}

    	return returnFields;
	}

	/**
	* Generate XML field data from a struct of data
	* @returnFields The structure of fields to generate data from
	*/
	private string function generateXMLData( required struct returnFields ){
		var buffer = createObject( "java", "java.lang.StringBuilder" ).init( '<?xml version="1.0" encoding="UTF-8"?><fields>' );

		for( var thisField in arguments.returnFields ){
			var key = xmlFormat( thisField );
			buffer.append( "<#key#>#xmlFormat( arguments.returnFields[ thisField ] )#</#key#>" );
		}

		buffer.append( "</fields>" );

		return buffer.toString();
	}

	/**
	* Populate Action PDF
	* @source The source PDF
	* @destination The destination PDF or if not passed, then stream to the browser
	* @flatten Flatten the results or not
	* @overwriteData Overwrite form data or not
	* @xmlData Use an XML source for data
	* @structData Use a structure for data
	* @jsonData Use a JSON object for data
	*
	* @return This instance
	*/
	private any function _populate(
		required any source,
		string destination,
		boolean flatten=true,
		boolean overwriteData=false,
		string xmlData,
		struct structData,
		string jsonData
	){
		// validate source
		if ( !fileExists( arguments.source ) ){
			throw( type="OrtusPDF.FileNotFound", message="The file '#arguments.source#' could not be found" );
		}

		// temp location according to destination
		if( !structKeyExists( arguments, "destination" ) || !len( arguments.destination ) ){
			var tempDestination = getTempDirectory() & "/" & createUUID() & ".tmp";
		} else {
			var tempDestination = arguments.destination & ".tmp";
		}
		// Are we populating from XML data
		if( !isNull( arguments.xmlData ) ){
			variables.formParams = getXMLFieldsAsStruct( arguments.xmlData );
		}
		// Are we populating from struct data
		else if( !isNull( arguments.structData ) ){
			variables.formParams = arguments.structData;
		}
		// Are we populating from struct data
		else if( !isNull( arguments.jsonData ) ){
			variables.formParams = deserializeJSON( arguments.jsonData );
		}

		// Get source fields
		var currentFieldsSet = getPDFFormFields( arguments.source );

		// java helpers
		var fileInputStream 	= createObject( "java", "java.io.FileInputStream" ).init( arguments.source );
		var PDFReader 			= createObject( "java", "org.apache.pdfbox.pdmodel.PDDocument", variables.LIB_PATH ).load( fileInputStream );
		var fileOutputStream 	= createObject( "java", "java.io.FileOutputStream" ).init( tempDestination );

		try{
			// Get catalog and acro fields
			var catalog 	= PDFReader.getDocumentCatalog();
			var acroForm 	= catalog.getAcroForm();

			// Iterate and do population of incoming form params
	        for( var name in variables.formParams ){
				var value = variables.formParams[ name ];
		        // get field
				var oField = acroForm.getField( name );
		        if( !isNull( oField ) ) {
		        	// Check if field has value and with overwritedata
		        	if( arguments.overwriteData || !len( oField.getValueAsString() ) ){
						oField.setValue( value );
					}
		        } else {
					// Ignore fields that do not exist instead of an exception
					writeDump( var="No field found with name: #name# when populating the pdf form. Current fields are #structKeyList( currentFieldsSet )#", output="console" );
		        }
	        }

	        // do we flatten
	        if( arguments.flatten ){
	        	acroForm.flatten();
	        }

	        // Save to temp destination
	        PDFReader.save( fileOutputStream );

	        // Verify destination or stream
	        if( !structKeyExists( arguments, "destination" ) || !len( arguments.destination ) ){
	        	cfcontent( type = "application/pdf", file = tempDestination, deleteFile = true );
	        } else {
	        	fileCopy( tempDestination, arguments.destination );
	        	fileDelete( tempDestination );
	        }
		} catch( any e ){
			rethrow;
		} finally {
			// Close it up
			fileInputStream.close();
			fileOutputStream.close();
			PDFReader.close();
		}

		return this;
	}

	/**
	* Read an xml file and return the fields as a struct of name-value pairs
	* @source The XML object, or XML string or XML data filename
	*/
	private struct function getXMLFieldsAsStruct( required string source ){
		var data = "";

		// XML string? or File Location?
		if( isXML( arguments.source ) || fileExists( arguments.source ) ){
			data = xmlParse( arguments.source );
		}
		// XML Doc
		else if( isXMLDoc( arguments.source ) ){
			data = arguments.source;
		}

		// Check for fields
		if( !structKeyExists( data, "fields" ) ){
			throw( type="OrtusPDF.FieldsNotFound", message="The XML data file '#arguments.source#' does not contain a <fields> item" );
		}

		var returnData = {};
		for( var thisChild in data.fields.XMLChildren ){
			returnData[ thisChild.xmlName ] = thisChild.xmlText;
		}
		return returnData;
	}

}